# Use standard logging in this module.
import logging

# Class managing list of things.
from dls_utilpack.things import Things

# Exceptions.
from dls_servbase_api.exceptions import NotFound

logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------------------
__default_cookie = None


def cookies_set_default(cookie):
    global __default_cookie
    __default_cookie = cookie


def cookies_get_default():
    global __default_cookie
    if __default_cookie is None:
        raise RuntimeError("cookies_get_default instance is None")
    return __default_cookie


# -----------------------------------------------------------------------------------------


class Cookies(Things):
    """
    List of available cookies.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, name=None):
        Things.__init__(self, name)

    # ----------------------------------------------------------------------------------------
    def build_object(self, specification, predefined_uuid=None):
        """"""

        cookie_class = self.lookup_class(specification["type"])

        try:
            cookie_object = cookie_class(specification, predefined_uuid=predefined_uuid)
        except Exception as exception:
            raise RuntimeError(
                "unable to build cookie object for type %s" % (cookie_class)
            ) from exception

        return cookie_object

    # ----------------------------------------------------------------------------------------
    def lookup_class(self, class_type):
        """"""

        if class_type == "dls_servbase_lib.cookies.dataface":
            from dls_servbase_lib.cookies.dataface import Dataface

            return Dataface

        raise NotFound("unable to get cookie class for type %s" % (class_type))



import asyncio
import copy
import json
import logging
import multiprocessing
import os
import socket
import threading
import time
import uuid
from urllib.parse import urlparse

import aiohttp
import aiohttp.web
import aiohttp.web_runner
from dls_logformatter.functions import list_exception_causes
from dls_mainiac_lib.mainiac import Mainiac
from dls_multiconf_lib.multiconfs import multiconfs_get_default, multiconfs_has_default
from dls_utilpack.callsign import callsign
from dls_utilpack.explain import explain
from dls_utilpack.global_signals import global_sigint
from dls_utilpack.modify_process_title import modify_process_title
from dls_utilpack.qualname import qualname
from dls_utilpack.search_file import SearchFileNotFound, search_file

from dls_servbase_api.aiohttp_client import AiohttpClient  # noqa: I001
from dls_servbase_api.constants import Keywords  # noqa: I001
from dls_servbase_lib.cookies.cookies import Cookies

logger = logging.getLogger(__name__)


class Opaque:
    def __init__(self):
        self.cookies = Cookies(name="http request cookies")
        pass


# ------------------------------------------------------------------------------------------
class BaseAiohttp:
    """
    Object representing a a process which receives requests from aiohttp.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, aiohttp_specification, calling_file=None):
        self.__aiohttp_specification = copy.deepcopy(aiohttp_specification)

        # Make sure there is a search_paths in the spedcification.
        search_paths = self.__aiohttp_specification.get("search_paths")
        if search_paths is None:
            self.__aiohttp_specification["search_paths"] = []

        # Caller's filename is presented?
        if calling_file is not None:
            # Search in html in the caller's filename's parent.
            self.__aiohttp_specification["search_paths"].append(
                f"{os.path.dirname(calling_file)}/html"
            )

        if "server" not in self.__aiohttp_specification:
            self.__aiohttp_specification["server"] = "http://%s:%s" % (
                self.__aiohttp_specification.get("server_host", "unconfigured"),
                self.__aiohttp_specification.get("port", "unconfigured"),
            )

        self.__callsign_url = self.__aiohttp_specification.get("server")
        if self.__callsign_url is None:
            self.__callsign_url = "http://%s:%s" % (
                self.__aiohttp_specification.get("server_host", "unconfigured"),
                self.__aiohttp_specification.get("port", "unconfigured"),
            )

        self.__callsign_url = self.__callsign_url.replace("*", socket.gethostname())

        # Build a client which can talk to this server.
        self.__aiohttp_client = AiohttpClient(self.__aiohttp_specification)

        self.__process2 = None
        self.__up_event2 = multiprocessing.Event()
        self.__up_event3 = asyncio.Event()

        # In the case where we are started as a process, we will use an new event loop of our own.
        self.owned_event_loop2 = None

        self.__thread3 = None

        # Event set when shutdown requested by remote request.
        self._shutdown_event = asyncio.Event()

        self.__thread = None
        self.__app_runner = None

        # Set log level default to avoid unwanted messages.
        logging.getLogger("aiohttp").setLevel(
            aiohttp_specification.get("log_level", "WARNING")
        )

        # We provide cookie services to our deriving classes.
        if "cookie_specification" in self.__aiohttp_specification:
            self.__cookie_specification = self.__aiohttp_specification[
                "cookie_specification"
            ]
        else:
            self.__cookie_specification = None

        # Some things for the health report.
        self.__protocolj_count = 0
        self.__file_count = 0
        self.__filestore_count = 0
        self.__time_started = time.time()

    # ----------------------------------------------------------------------------------------
    def callsign(self):
        """"""
        return self.__callsign_url

    # ----------------------------------------------------------------------------------------
    async def start_process(self):
        """"""

        self.__process2 = multiprocessing.Process(
            target=self.prepare_to_activate_process,
        )
        self.__process2.start()
        logger.debug(f"[PIDAL] {callsign(self)} pid {self.__process2.pid} is started")
        timeout = 5.0
        naptime = 0.25
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self.__up_event2.is_set():
                return
            if not self.__process2.is_alive():
                raise RuntimeError(
                    f"process for {callsign(self)} died without setting up_event"
                )
            await asyncio.sleep(naptime)

        raise RuntimeError(
            f"process for {callsign(self)} did not set up_event within timeout {timeout} seconds"
        )

    # ----------------------------------------------------------------------------------------
    async def is_process_started(self):
        """"""

        return self.__process2 is not None

    # ----------------------------------------------------------------------------------------
    async def is_process_alive(self):
        """"""

        if self.__process2 is None:
            raise RuntimeError(f"{callsign(self)} has no process defined for it")

        is_alive = self.__process2.is_alive()

        # state = "alive" if is_alive else "dead"

        # logger.debug(f"[PIDAL] {callsign(self)} pid {self.__process2.pid} is {state}")

        return is_alive

    # ----------------------------------------------------------------------------------------
    def prepare_to_activate_process(self):
        """"""

        try:
            if multiconfs_has_default():
                dls_servbase_multiconf = multiconfs_get_default()

                # ---------------------------------------------------------------------
                # Multiconf must provide logging settings.
                logging_settings = dls_servbase_multiconf.require("logging_settings")

                # Remove existing handlers which may have been inherited from a fork.
                handlers = list(logging.getLogger().handlers)
                for handler in handlers:
                    logging.getLogger().removeHandler(handler)

                # Make a dummy mainiac and use it to configure logging.
                # Some loggers need to connect to logging servers in the new process.
                mainiac = Mainiac(callsign(self))
                mainiac.configure_logging(logging_settings)

                # ---------------------------------------------------------------------
                # We need special method for mpqueue since it is a multiprocessing object.
                # logging_mpqueue = dls_servbase_multiconf.get_logging_mpqueue()
                # if logging_mpqueue is not None:
                #     # Remove existing handlers.
                #     handlers = list(logging.getLogger().handlers)
                #     for handler in handlers:
                #         logging.getLogger().removeHandler(handler)
                #     mpqueue_handler = logging.handlers.QueueHandler(logging_mpqueue)
                #     mpqueue_handler.setLevel(logging.DEBUG)
                #     # Let logging write custom formatted messages to stdout.
                #     # This will format the full stack trace in the current interpreter.
                #     # So the queue listener has no opportunity to prefer a format like "bare" or "short".
                #     # TODO: Improve logform formatting options when using mpqueue.
                #     mpqueue_handler.setFormatter(DlsLogformatter(type="long"))
                #     logging.getLogger().handlers = []
                #     logging.getLogger().addHandler(mpqueue_handler)
                # else:
                #     logger.debug("[MPQLOG] no mpqueue")
            else:
                logger.debug("[MPQLOG] no multiconf")

        except Exception as exception:
            logger.exception(
                callsign(
                    self,
                    explain(exception, "preparing to activate process"),
                ),
                exc_info=exception,
            )

        # Call abstract method provided by the overriding class.
        self.activate_process()

    # ----------------------------------------------------------------------------------------
    def activate_process_base(self):
        """"""

        try:
            # Change our title as it appears in "ps ax" and "top".
            modify_process_title(callsign(self))

            # Activate the signal handling.
            global_sigint.activate()

            self.owned_event_loop2 = asyncio.new_event_loop()
            asyncio.set_event_loop(self.owned_event_loop2)
            coro_future = self.activate_coro()
            self.owned_event_loop2.run_until_complete(coro_future)

            # Now we got a server running we can allow the parent process to continue.
            self.__up_event2.set()

            self.owned_event_loop2.run_forever()

        except Exception as exception:
            logger.exception("exception in process", exc_info=exception)

        logger.debug(f"[PIDAL] {callsign(self)} pid {os.getpid()} has quit")

    # ----------------------------------------------------------------------------------------
    async def start_thread(self):
        """"""

        loop = asyncio.get_event_loop()
        self.__thread3 = threading.Thread(target=self.activate_thread, args=(loop,))
        self.__thread3.start()
        await self.__up_event3.wait()

    # ----------------------------------------------------------------------------------------
    def activate_thread_base(self, loop):
        """
        This is called from each server's activate_thread() method.
        """

        try:
            # Activate the server.
            future = asyncio.run_coroutine_threadsafe(self.activate_coro(), loop)

            # Wait for the activation to complete.
            future.result()

            # Now we got a server running we can allow the caller to continue.
            loop.call_soon_threadsafe(self.__up_event3.set)

        except Exception as exception:
            logger.exception("exception in thread", exc_info=exception)

    # ----------------------------------------------------------------------------------------
    async def activate_coro_base(self, route_tuples=[]):
        """
        This is called from each server's activate_coro() method.
        """
        try:
            # Shut down any existing servers.
            result = await self.client_shutdown()

            if result.get("had_existed", False):
                logger.info(f"we have shut down an existing process {callsign(self)}")
            else:
                logger.debug(f"[AUTSHUT] {callsign(self)} {result['message']}")

            web_app = aiohttp.web.Application()

            # TODO: Make the route_tuples in base_aiohttp specific to each server in question.
            routes = [
                aiohttp.web.get("/shutdown", self._route_shutdown),
                aiohttp.web.get("/report_health", self._route_report_health),
                aiohttp.web.post("/protocolj", self._route_protocolj),
                aiohttp.web.get(
                    "/filestore/{filestore_path:.*}", self._route_get_filestore
                ),
                aiohttp.web.get("/{url:.*}", self._route_get_file),
            ]

            # Caller can give extra routes like this:
            # route_tuples = [("post", "submit", self._route_submit)]

            for route_tuple in route_tuples:
                if route_tuple[0] == "get":
                    routes.append(
                        aiohttp.web.get("/%s" % (route_tuple[1]), route_tuple[2])
                    )
                elif route_tuple[0] == "post":
                    routes.append(
                        aiohttp.web.post("/%s" % (route_tuple[1]), route_tuple[2])
                    )

            web_app.add_routes(routes)

            self.__app_runner = aiohttp.web.AppRunner(web_app)

            await self.__app_runner.setup()

            endpoint = self.__aiohttp_specification["server"]
            if endpoint.startswith("http"):
                parts = urlparse(endpoint)
                site = aiohttp.web.TCPSite(
                    self.__app_runner, parts.hostname, parts.port
                )
            else:
                parts = urlparse(endpoint)
                site = aiohttp.web.UnixSite(self.__app_runner, parts.path)

            await site.start()
        except Exception as exception:
            raise RuntimeError(
                explain(exception, f"exception starting {callsign(self)} site")
            )

    # ----------------------------------------------------------------------------------------
    async def base_direct_shutdown(self):
        """"""
        # If we were started via process, we will need to stop our own event loop.
        if self.owned_event_loop2 is not None:
            logger.debug(
                f"[NEWSHUT] directly shutting down {callsign(self)} owned event loop on pid {os.getpid()}"
            )

            try:
                self.owned_event_loop2.stop()

            except Exception as exception:
                raise RuntimeError(callsign(self, explain(exception, "shutting down")))

        elif self.__app_runner is not None:
            logger.debug(
                f"[NEWSHUT] directly shutting down {callsign(self)} by awaiting self.__app_runner.cleanup()"
            )
            await self.__app_runner.cleanup()

        # Set event which will cause server to shutdown.
        self._shutdown_event.set()

    # ----------------------------------------------------------------------------------------
    async def wait_for_shutdown(self):
        """"""
        try:
            await self._shutdown_event.wait()

        except Exception as exception:
            raise RuntimeError("exception while waiting for shutdown") from exception

    # ------------------------------------------------------------------------------------------
    # Handle request to get static file.
    async def _route_get_file(self, request):
        try:
            # Stuff for the health report.
            self.__file_count += 1

            # logger.info(f"request.path is {request.path}")

            # Filename from the url line.
            filename = request.path

            search_paths = self.__aiohttp_specification.get("search_paths", [])
            try:
                absolute = search_file(search_paths, filename)
            except SearchFileNotFound as exception:
                response = aiohttp.web.Response(
                    status=404, reason=str(exception), text=str(exception)
                )
                return response

            suffix = os.path.splitext(absolute)[1]

            if suffix == ".html":
                content_type = "text/html"
            elif suffix == ".css":
                content_type = "text/css"
            elif suffix == ".js":
                content_type = "application/javascript"
            elif suffix == ".json":
                content_type = "application/json"
            elif suffix == ".png":
                content_type = "image/png"
            elif suffix == ".jpg":
                content_type = "image/jpg"
            else:
                content_type = "application/octet-stream"

            headers = {"Content-Type": content_type}

            response = aiohttp.web.FileResponse(absolute, headers=headers)

            return response
        except Exception as exception:
            return self._compose_standard_exception(exception)

    # ------------------------------------------------------------------------------------------
    # Handle request to get file from filestore.
    async def _route_get_filestore(self, request):
        try:
            # Stuff for the health report.
            self.__filestore_count += 1

            absolute = "/" + request.match_info["filestore_path"].strip("/")

            suffix = os.path.splitext(absolute)[1]

            if suffix == ".html":
                content_type = "text/html; charset=utf-8"
            elif suffix == ".log":
                content_type = "text/plain; charset=utf-8"
            elif suffix == ".stdout":
                content_type = "text/plain; charset=utf-8"
            elif suffix == ".stderr":
                content_type = "text/plain; charset=utf-8"
            elif suffix == ".txt":
                content_type = "text/plain; charset=utf-8"
            elif suffix == ".css":
                content_type = "text/css; charset=utf-8"
            elif suffix == ".csv":
                content_type = "text/csv; charset=utf-8"
            elif suffix == ".yml" or suffix == ".yaml":
                content_type = "application/x-yaml"
            elif suffix == ".json":
                content_type = "application/json"
            elif suffix == ".png":
                content_type = "image/png"
            elif suffix == ".gif":
                content_type = "image/gif"
            elif suffix == ".jpg":
                content_type = "image/jpg"
            else:
                content_type = "application/octet-stream"

            headers = {"Content-Type": content_type}

            response = aiohttp.web.FileResponse(absolute, headers=headers)

            return response
        except Exception as exception:
            return self._compose_standard_exception(exception)

    # ------------------------------------------------------------------------------------------
    # Handle generic protocolj request.
    async def _route_protocolj(self, web_request):
        try:
            # Stuff for the health report.
            self.__protocolj_count += 1

            # Make an object to identify this transaction.
            opaque = Opaque()

            # Read the full body of the request and parse as json.
            content = await web_request.json()

            await self.register_cookies(
                opaque, web_request, content.get(Keywords.ENABLE_COOKIES, [])
            )

            # Dispatch the request to be handled per business logic.
            response_dict = await self.dispatch(content, opaque)

            # Make a response object.
            web_response = aiohttp.web.json_response(response_dict)

            # Flush all cookies.
            await self.flush_cookies(opaque, web_response)

            return web_response
        except Exception as exception:
            return self._compose_standard_exception(exception)

        finally:
            # Release all registered cookies to free their connections and other resources.
            await self.release_cookies(opaque)

    # ----------------------------------------------------------------------------------------
    async def register_cookies(self, opaque, web_request, cookie_names):
        """
        Register all the given named cookies received in the web request matching given names.
        """

        # Configuration does not provide a cookie specification?
        if self.__cookie_specification is None:
            # Just don't do cookies.
            # Other calls may fail, alerting the developer to the mis-configuration.
            return

        logger.debug(f"[COOKOFF] registering cookies {cookie_names}")

        # The caller gives the list of cookies that are interesting.
        for cookie_name in cookie_names:
            # We have a uuid for this cookie name already?
            if cookie_name in web_request.cookies:
                _uuid = web_request.cookies[cookie_name]
                logger.debug(
                    f"[COOKOFF] registering cookie {cookie_name} from existing uuid {_uuid}"
                )
                if _uuid == "" or _uuid == "None":
                    _uuid = None
            else:
                _uuid = None

            # There is no cookie by this name yet?
            if _uuid is None:
                _uuid = str(uuid.uuid4())
                logger.debug(
                    f"[COOKOFF] registering cookie {cookie_name} gets new uuid {_uuid}"
                )

            specification = copy.deepcopy(self.__cookie_specification)
            cookie = opaque.cookies.build_object(specification, predefined_uuid=_uuid)
            cookie.traits()["cookie_name"] = cookie_name

            opaque.cookies.add(cookie)

    # ----------------------------------------------------------------------------------------
    async def release_cookies(self, opaque):
        """
        Release all registered cookies to free their connections and other resources.
        """

        if self.__cookie_specification is None:
            return

        logger.debug("[RELCOOK] releasing cookies")

        for cookie in opaque.cookies.list():
            logger.debug("[RELCOOK] releasing cookie")
            await cookie.release()

    # ----------------------------------------------------------------------------------------
    async def flush_cookies(self, opaque, web_response):
        """
        Add all registered cookies to the response and save their persisted values.
        """

        if self.__cookie_specification is None:
            return

        # logger.debug("[COOKOFF] saving cookies to persist")

        for cookie in opaque.cookies.list():
            # Put the cookie into the web response.
            cookie_name = cookie.trait("cookie_name")
            web_response.set_cookie(cookie_name, cookie.uuid(), max_age=10000000)

            if cookie.is_dirty():
                # logger.debug(f"[COOKOFF] saving {cookie_name} to persist")
                # Store the cookie persistently.
                await cookie.save_to_persist()

    # ----------------------------------------------------------------------------------------
    def get_cookie_contents(self, opaque, cookie_name):
        """"""

        if self.__cookie_specification is None:
            raise RuntimeError("cookies have not been configured")

        cookie = opaque.cookies.find(cookie_name, trait_name="cookie_name")

        return cookie.get_contents()

    # ----------------------------------------------------------------------------------------
    async def get_cookie_content(
        self, opaque, cookie_name, keyword, default=None, xfrom=None
    ):
        """"""

        # No cookie specification in the configuration file?
        if self.__cookie_specification is None:
            raise RuntimeError("cookies have not been configured")

        # Find the cookie object corresponding to the cookie's name.
        cookie = opaque.cookies.find(cookie_name, trait_name="cookie_name")

        # Cookie has not been set yet?
        if not cookie.has_contents():
            # Get the cookie from the persistent storage.
            logger.debug(
                f"[COOKSEL] from {xfrom} loading cookie {id(cookie)} {cookie.uuid()} {cookie_name} from persist"
            )
            await cookie.load_from_persist()

        # Get the contents (dictionary) of the cookie.
        contents = cookie.get_contents()

        # Get the keyword from within the cookie contents dict.
        value = contents.get(keyword, default)

        return value

    # ----------------------------------------------------------------------------------------
    def set_cookie_content(self, opaque, cookie_name, keyword, value):
        """"""

        if self.__cookie_specification is None:
            raise RuntimeError("cookies have not been configured")

        cookie = opaque.cookies.find(cookie_name, trait_name="cookie_name")

        if cookie.has_contents():
            contents = cookie.get_contents()
        else:
            contents = {}
            cookie.set_contents(contents)
            cookie.set_dirty()

        if keyword not in contents or contents[keyword] != value:
            contents[keyword] = value
            cookie.set_dirty()

    # ----------------------------------------------------------------------------------------
    async def set_or_get_cookie_content(
        self, opaque, cookie_name, keyword, new_value, default
    ):

        if new_value == "undefined":
            new_value = None

        # Request does not contain the auto_update_enabled state in the ux?
        # This happens at the first request after a page load.
        if new_value is None:
            new_value = await self.get_cookie_content(
                opaque,
                cookie_name,
                keyword,
                default=default,
                xfrom="set_or_get_cookie_content",
            )

        self.set_cookie_content(
            opaque,
            cookie_name,
            keyword,
            new_value,
        )

        return new_value

    # ----------------------------------------------------------------------------------------
    async def dispatch(self, request_dict):
        """"""

        # Parent class must override this method.
        raise RuntimeError(f"no dispatching available by {callsign(self)} server")

    # ------------------------------------------------------------------------------------------
    async def _route_report_health(self, request):
        try:
            report = await self.report_health()

            # Make a response object.
            web_response = aiohttp.web.json_response(report)

            return web_response
        except Exception as exception:
            return self._compose_standard_exception(exception)

    # ------------------------------------------------------------------------------------------
    # Report server health.
    async def report_health(self):

        report_aiohttp = {}
        report_aiohttp["protocolj_count"] = self.__protocolj_count
        report_aiohttp["file_count"] = self.__file_count
        report_aiohttp["filestore_count"] = self.__filestore_count

        report = {}
        report["aiohttp"] = report_aiohttp
        report["time_alive"] = "%0.1f" % (time.time() - self.__time_started)
        report["request_count"] = (
            self.__protocolj_count + self.__file_count + self.__filestore_count
        )

        return report

    # ------------------------------------------------------------------------------------------
    # Handle request to exit the process.
    async def _route_shutdown(self, request):
        try:
            # Schedule a shutdown to happen after we finish this response.
            asyncio.create_task(self.direct_shutdown())

            response = aiohttp.web.Response(status=200, text="ok")

            return response
        except Exception as exception:
            return self._compose_standard_exception(exception)

    # ----------------------------------------------------------------------------------------
    def _compose_standard_exception(self, exception):
        logger.exception(
            f"{callsign(self)} exception while handling a request: {str(exception)}",
            exc_info=exception,
        )

        exception_causes = list_exception_causes(exception)
        body = {
            "exception": {
                "qualname": qualname(exception),
                "message": str(exception),
                "exception_causes": exception_causes,
            }
        }

        return aiohttp.web.Response(status=500, body=json.dumps(body, indent=4))

    # ----------------------------------------------------------------------------------------
    def client_url(self):
        """"""
        return self.__callsign_url

    # ----------------------------------------------------------------------------------------
    async def client_protocolj(self, request_object, cookies=None):
        """"""
        return await self.__aiohttp_client.client_protocolj(
            request_object, cookies=cookies
        )

    # ----------------------------------------------------------------------------------------
    async def client_get_json(self, url):
        """"""
        return await self.__aiohttp_client.client_get_json(url)

    # ----------------------------------------------------------------------------------------
    async def client_get_csv(self, url):
        """"""
        return await self.__aiohttp_client.client_get_csv(url)

    # ----------------------------------------------------------------------------------------
    async def client_get_file(self, url):
        """"""
        return await self.__aiohttp_client.client_get_file(url)

    # ----------------------------------------------------------------------------------------
    async def client_shutdown(self):
        """"""

        return await self.__aiohttp_client.client_shutdown()

    # ----------------------------------------------------------------------------------------
    async def close_client_session(self):
        """"""
        return await self.__aiohttp_client.close_client_session()

    # ----------------------------------------------------------------------------------------
    async def client_report_health(self):
        """"""
        return await self.__aiohttp_client.client_report_health()



# Use standard logging in this module.
import logging

# Class managing list of things.
from dls_utilpack.things import Things

# Exceptions.
from dls_servbase_api.exceptions import NotFound

logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------------------
__default_dls_servbase_gui = None


def dls_servbase_guis_set_default(dls_servbase_gui):
    global __default_dls_servbase_gui
    __default_dls_servbase_gui = dls_servbase_gui


def dls_servbase_guis_get_default():
    global __default_dls_servbase_gui
    if __default_dls_servbase_gui is None:
        raise RuntimeError("dls_servbase_guis_get_default instance is None")
    return __default_dls_servbase_gui


def dls_servbase_guis_has_default():
    global __default_dls_servbase_gui
    return __default_dls_servbase_gui is not None


# -----------------------------------------------------------------------------------------


class Guis(Things):
    """
    List of available dls_servbase_guis.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, name=None):
        Things.__init__(self, name)

    # ----------------------------------------------------------------------------------------
    def build_object(self, specification):
        """"""

        dls_servbase_gui_class = self.lookup_class(specification["type"])

        try:
            dls_servbase_gui_object = dls_servbase_gui_class(specification)
        except Exception as exception:
            raise RuntimeError(
                "unable to build dls_servbase_gui object for type %s"
                % (dls_servbase_gui_class)
            ) from exception

        return dls_servbase_gui_object

    # ----------------------------------------------------------------------------------------
    def lookup_class(self, class_type):
        """"""

        if class_type == "dls_servbase_lib.dls_servbase_guis.aiohttp":
            from dls_servbase_lib.guis.aiohttp import Aiohttp

            return Aiohttp

        raise NotFound(
            "unable to get dls_servbase_gui class for type %s" % (class_type)
        )



import asyncio

# Use standard logging in this module.
import logging

# Utilities.
from dls_utilpack.callsign import callsign

# Base class for cli subcommands.
from dls_servbase_cli.subcommands.base import Base

# Context creator.
from dls_servbase_lib.contexts.contexts import Contexts

# Special reference to the gui so we give the url to the user on the info.
from dls_servbase_lib.guis.guis import dls_servbase_guis_get_default

logger = logging.getLogger()

# Specifications of services we can start, and their short names for parse args.
services = {
    "dls_servbase_dataface_specification": "dataface",
    "dls_servbase_collector_specification": "collector",
    "dls_servbase_gui_specification": "gui",
}


# --------------------------------------------------------------
class StartServices(Base):
    """
    Start one or more services and keep them running until ^C.
    """

    def __init__(self, args, mainiac):
        super().__init__(args)

        self.__mainiac = mainiac

    # ----------------------------------------------------------------------------------------
    def run(self):
        """ """

        # Run in asyncio event loop.
        asyncio.run(self.__run_coro())

    # ----------------------------------------------------------
    async def __run_coro(self):
        """"""

        # Load the configuration.
        dls_servbase_multiconf = self.get_multiconf()
        context_configuration = await dls_servbase_multiconf.load()

        if "all" in self._args.service_names:
            selected_service_names = []
            for _, service_name in services.items():
                selected_service_names.append(service_name)
        else:
            selected_service_names = self._args.service_names

        # Change all start_as to None, except the one we are starting.
        for keyword, specification in context_configuration.items():
            if keyword in services:
                service_name = services[keyword]
                if service_name in selected_service_names:
                    specification["context"] = {"start_as": "process"}

        # Make a context from the configuration.
        dls_servbase_context = Contexts().build_object(context_configuration)

        # Open the context (servers and clients).
        async with dls_servbase_context:
            if "gui" in selected_service_names:
                logger.info(f"starting gui {callsign(dls_servbase_guis_get_default())}")

            try:
                # Stay up until all processes are dead.
                # TODO: Use asyncio wait or sentinel for all started processes to be dead.
                while True:
                    await asyncio.sleep(1.0)
                    if not await dls_servbase_context.is_any_process_alive():
                        logger.info("all processes have shutdown")
                        break

            except KeyboardInterrupt:
                pass

    # ----------------------------------------------------------
    def add_arguments(parser):

        services_list = list(services.values())

        parser.add_argument(
            help='"all" or any combination of {%s}' % (" ".join(services_list)),
            nargs="+",
            type=str,
            metavar="service name(s)",
            dest="service_names",
        )

        return parser



import json
import logging

# Utilities.
from dls_utilpack.callsign import callsign
from dls_utilpack.require import require

# Dataface interface.
from dls_servbase_api.datafaces.datafaces import Datafaces

# Base class for cookie things.
from dls_servbase_lib.cookies.base import Base as CookieBase

logger = logging.getLogger(__name__)

thing_type = "dls_servbase_lib.cookies.dataface"


class Dataface(CookieBase):
    """ """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None, predefined_uuid=None):
        CookieBase.__init__(
            self, thing_type, specification, predefined_uuid=predefined_uuid
        )

        self.__type_specific_tbd = require(
            f"{callsign(self)} specification", self.specification(), "type_specific_tbd"
        )

        dataface_specification = require(
            f"{callsign(self)} specification root",
            self.__type_specific_tbd,
            "dataface_specification",
        )

        self.__dataface = Datafaces().build_object(dataface_specification)

    # ----------------------------------------------------------------------------------------
    async def load_from_persist(self):
        """
        Load cookie contents from persistent storage.
        """

        if self.__dataface is None:
            raise RuntimeError("disallowed attempt to reload from persist")

        record = await self.__dataface.get_cookie(self.uuid())

        if record is None:
            self.set_contents({})
        else:
            self.set_contents(json.loads(record["contents"]))

    # ----------------------------------------------------------------------------------------
    async def save_to_persist(self):
        if self.has_contents():
            contents = self.get_contents()
        else:
            contents = {}

        record = await self.__dataface.get_cookie(self.uuid())
        # logger.debug(describe("[COOKOFF] saving cookie record", record))

        if record is None:
            # logger.debug(
            #     f"[COOKOFF] creating cookie record for {self.uuid()} {contents}"
            # )
            record = {
                "uuid": self.uuid(),
                "name": self.traits()["cookie_name"],
                "contents": json.dumps(contents),
            }
            await self.__dataface.set_cookie(record)
        else:
            record = {"uuid": self.uuid(), "contents": json.dumps(contents)}
            # logger.debug(
            #     f"[COOKOFF] updating cookie record for {self.uuid()} {contents}"
            # )
            await self.__dataface.update_cookie(record)

    # ----------------------------------------------------------------------------------------
    async def release(self):
        """
        Release cookie and free its connections and other resources.
        """

        logger.debug(
            f"[RELCOOK] releasing client to cookie dataface {str(self.__dataface)}"
        )
        if self.__dataface is not None:
            await self.__dataface.close_client_session()
            self.__dataface = None



import logging

# Base class for the aiosqlite database object.
from dls_normsql.aiosqlite import Aiosqlite as NormsqlAiosqlite

# Base class for our database definition.
from dls_servbase_lib.databases.database_definition import DatabaseDefinition

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class Aiosqlite(DatabaseDefinition, NormsqlAiosqlite):
    """
    Class with coroutines for creating and querying a sqlite database.
    We use dls-normsql to do the heavy lifting.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        """
        Construct object.  Do not connect to database.
        """

        # Constructor for the database definition.
        DatabaseDefinition.__init__(self)

        # Constructor for the database implementation.
        NormsqlAiosqlite.__init__(self, specification)

    # ----------------------------------------------------------------------------------------
    def reinstance(self):
        """
        Clone database instance.
        This is needed if a process inherits an instance, but needs its own connection.
        """

        return Aiosqlite(self.__filename)

    # ----------------------------------------------------------------------------------------
    async def add_table_definitions(self):
        """
        Make all the table definitions.
        """

        # Add tables common in all implementations.
        await NormsqlAiosqlite.add_table_definitions(self)

        # Add tables from our definition.
        await DatabaseDefinition.add_table_definitions(self)



import logging
import multiprocessing
import threading

# Utilities.
from dls_utilpack.callsign import callsign
from dls_utilpack.require import require

# Basic things.
from dls_utilpack.thing import Thing

# Base class for an aiohttp server.
from dls_servbase_lib.base_aiohttp import BaseAiohttp

# Gui protocolj things (must agree with javascript).
from dls_servbase_lib.guis.constants import Commands, Cookies, Keywords

logger = logging.getLogger(__name__)

thing_type = "dls_servbase_lib.dls_servbase_guis.aiohttp"


# ------------------------------------------------------------------------------------------
class Aiohttp(Thing, BaseAiohttp):
    """
    Object implementing remote procedure calls for dls_servbase_gui methods.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):
        Thing.__init__(self, thing_type, specification)
        BaseAiohttp.__init__(
            self,
            specification["type_specific_tbd"]["aiohttp_specification"],
            calling_file=__file__,
        )

    # ----------------------------------------------------------------------------------------
    def callsign(self):
        """"""
        return "%s %s" % ("Gui.Aiohttp", BaseAiohttp.callsign(self))

    # ----------------------------------------------------------------------------------------
    def activate_process(self):
        """"""

        try:
            multiprocessing.current_process().name = "gui"

            self.activate_process_base()

        except Exception as exception:
            logger.exception(
                f"unable to start {callsign(self)} process", exc_info=exception
            )

    # ----------------------------------------------------------------------------------------
    def activate_thread(self, loop):
        """
        Called from inside a newly created thread.
        """

        try:
            threading.current_thread().name = "gui"

            self.activate_thread_base(loop)

        except Exception as exception:
            logger.exception(
                f"unable to start {callsign(self)} thread", exc_info=exception
            )

    # ----------------------------------------------------------------------------------------
    async def activate_coro(self):
        """"""
        try:
            # No special routes, we will use protocolj dispathcing only
            route_tuples = []

            # Start the actual coro to listen for incoming http requests.
            await self.activate_coro_base(route_tuples)

        except Exception:
            raise RuntimeError(f"unable to start {callsign(self)} server coro")

    # ----------------------------------------------------------------------------------------
    async def direct_shutdown(self):
        """"""

        # Let the base class stop the server looping.
        await self.base_direct_shutdown()

    # ----------------------------------------------------------------------------------------
    async def dispatch(self, request_dict, opaque):
        """"""

        command = require("request json", request_dict, Keywords.COMMAND)

        if command == Commands.LOAD_TABS:
            return await self._load_tabs(opaque, request_dict)

        if command == Commands.SELECT_TAB:
            return await self._select_tab(opaque, request_dict)

        else:
            raise RuntimeError("invalid command %s" % (command))

    # ----------------------------------------------------------------------------------------
    async def _load_tabs(self, opaque, request_dict):

        tab_id = await self.get_cookie_content(
            opaque,
            Cookies.TABS_MANAGER,
            Keywords.TAB_ID,
            xfrom="guis.Aiohttp_load_tabs",
        )
        logger.debug(f"[GUITABS] tab_id from cookie content is {tab_id}")

        # Reply with tabs.
        response = {Keywords.TAB_ID: tab_id}

        return response

    # ----------------------------------------------------------------------------------------
    async def _select_tab(self, opaque, request_dict):
        tab_id = require("request json", request_dict, Keywords.TAB_ID)

        logger.debug(f"[GUITABS] tab_id in request is {tab_id}")

        # Put the tab_id into the cookie.
        self.set_cookie_content(opaque, Cookies.TABS_MANAGER, Keywords.TAB_ID, tab_id)

        response = {}

        return response



class Keywords:
    COMMAND = "dls_servbase_guis::keywords::command"
    PAYLOAD = "dls_servbase_guis::keywords::payload"
    TAB_ID = "dls_servbase_guis::keywords::tab_id"


class Commands:
    LOAD_TABS = "dls_servbase_guis::commands::load_tabs"
    SELECT_TAB = "dls_servbase_guis::commands::select_tab"


class Cookies:
    TABS_MANAGER = "dls_servbase_guis::cookies::tabs_manager"



import asyncio
import copy
import json
import logging
import socket
import time
from urllib.parse import urlparse

import aiohttp

# Utilities.
from dls_utilpack.callsign import callsign
from dls_utilpack.describe import describe
from dls_utilpack.explain import explain, explain2

# Exceptions.
from dls_servbase_api.exceptions import (
    ClientConnectorError as DlsServbaseClientConnectorError,
)

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------------------
class AiohttpClient:
    """
    Object representing a client to an aiohttp server.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, aiohttp_specification):
        self.__aiohttp_specification = copy.deepcopy(aiohttp_specification)

        if "client" not in self.__aiohttp_specification:
            self.__aiohttp_specification["client"] = "http://%s:%s" % (
                self.__aiohttp_specification.get("client_host", "unconfigured"),
                self.__aiohttp_specification.get("port", "unconfigured"),
            )

        self.__endpoint = self.__aiohttp_specification["client"]
        self._client_session = None

        # Set log level default to avoid unwanted messages.
        logging.getLogger("aiohttp").setLevel(
            aiohttp_specification.get("log_level", "WARNING")
        )

    # ----------------------------------------------------------------------------------------
    def callsign(self):
        """"""
        return self.__aiohttp_specification["client"]

    # ----------------------------------------------------------------------------------------
    async def __is_client_connection_possible(self):
        """"""

        if self.__endpoint.startswith("http"):
            parts = urlparse(self.__endpoint)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = sock.connect_ex((parts.hostname, parts.port))
            sock.close()
            # logger.debug(
            #     f"sock.connect_ex(({parts.hostname}, {parts.port})) result is {result}"
            # )
            return result == 0
        else:
            # TODO: Do proper check if client connection possible for unix pipe.
            await asyncio.sleep(0.100)
            return True

    # ----------------------------------------------------------------------------------------
    async def _establish_client_session(self):
        """"""

        if self._client_session is None:
            if self.__endpoint.startswith("http"):
                connector = None
                self.__client_netloc = self.__endpoint
            else:
                parts = urlparse(self.__endpoint)
                connector = aiohttp.UnixConnector(path=parts.path)
                self.__client_netloc = "http://unixconnector"

            self._client_session = aiohttp.ClientSession(connector=connector)

    # ----------------------------------------------------------------------------------------
    async def client_get_info(self):
        """"""
        await self._establish_client_session()

        async with self._client_session.get("/get_info") as response:
            return await response.json()

    # ----------------------------------------------------------------------------------------
    async def client_protocolj(self, request_object, cookies=None):
        """"""
        await self._establish_client_session()

        # logger.info(describe("request_object", request_object))

        try:
            async with self._client_session.post(
                f"{self.__client_netloc}/protocolj",
                json=request_object,
                cookies=cookies,
            ) as response:
                if response.status == 200:
                    response_json = await response.json()
                    if cookies is not None:
                        if response_json is None:
                            response_json = {}
                        response_json["__cookies"] = response.cookies
                    return response_json
                else:
                    text = await response.text()

                    exception_class = RuntimeError
                    exception_message = "server responded with status %d: %s" % (
                        response.status,
                        text,
                    )

                    try:
                        response_dict = json.loads(text)

                        if "exception" in response_dict:
                            qualname = response_dict["exception"].get("qualname")
                            exception_message = response_dict["exception"].get(
                                "message", "no message"
                            )
                            exception_class = RuntimeError
                            exception_message = f"{qualname}: {exception_message}"
                    except Exception:
                        pass

                    raise exception_class(exception_message)
        except aiohttp.ClientConnectorError as exception:
            raise DlsServbaseClientConnectorError(
                explain(exception, f"connecting to {callsign(self)}")
            )

    # ----------------------------------------------------------------------------------------
    def _set_auth_headers(self, headers):
        """
        Put the proper authorization in the headers.
        """

        # Already authorization there?
        if "Authorization" in headers:
            return

        config = self.__aiohttp_specification.get("authorization")
        if config is None:
            return

        type = config.get("type")
        if type != "http_basic":
            raise RuntimeError(
                f'{callsign(self)} cannot understand configured authorization type "{type}"'
            )

        username = config.get("username")
        if username is None:
            raise RuntimeError(
                f"{callsign(self)} configured authorization does not contain username"
            )
        password = config.get("password")
        if password is None:
            raise RuntimeError(
                f"{callsign(self)} caonfigured authorization does not contain password"
            )

        auth = aiohttp.BasicAuth(username, password)

        headers["Authorization"] = auth.encode()

    # ----------------------------------------------------------------------------------------
    async def client_get_json(self, url, **kwargs):
        """"""
        await self._establish_client_session()

        if not url.startswith("/"):
            url = "/" + url
        url = f"{self.__client_netloc}{url}"

        headers = kwargs.get("headers", {})
        headers = copy.deepcopy(headers)
        headers["Accept"] = "application/json"

        # Put the proper authorization in the headers.
        self._set_auth_headers(headers)

        kwargs["headers"] = headers

        # logger.debug(describe(f"{callsign(self)} kwargs", kwargs))

        async with self._client_session.get(url, **kwargs) as response:

            if response.status != 200:
                content = await response.text()
                raise RuntimeError(
                    f"{url} error {response.status}: {response.reason}\n{content}"
                )

            content = await response.json()

            return content

    # ----------------------------------------------------------------------------------------
    async def client_get_csv(self, url, **kwargs):
        """"""
        await self._establish_client_session()

        if not url.startswith("/"):
            url = "/" + url
        url = f"{self.__client_netloc}{url}"

        headers = kwargs.get("headers", {})
        headers = copy.deepcopy(headers)
        headers["Accept"] = "text/csv"
        kwargs["headers"] = headers
        logger.debug(describe("kwargs", kwargs))

        async with self._client_session.get(url, **kwargs) as response:
            logger.info(f"response.url {response.url}")

            if response.status != 200:
                content = await response.text()
                raise RuntimeError(
                    f"{url} error {response.status}: {response.reason}\n{content}"
                )

            content = await response.text()

            return content

    # ----------------------------------------------------------------------------------------
    async def client_get_file(self, url):
        """"""
        await self._establish_client_session()

        if not url.startswith("/"):
            url = "/" + url
        url = f"{self.__client_netloc}{url}"

        async with self._client_session.get(url) as response:

            if response.status != 200:
                raise RuntimeError(f"{url} error {response.status}: {response.reason}")

            if response.headers["Content-Type"] in (
                "text/plain"
                "text/html"
                "text/css"
                "application/javascript"
                "application/json"
            ):

                # Read body and decode into text.
                content = await response.text("utf-8")
            else:
                content = await response.read()

            return content

    # ----------------------------------------------------------------------------------------
    async def client_report_health(self):
        """"""
        await self._establish_client_session()

        try:
            response = await self.client_get_json("report_health")
        except Exception as exception:
            logger.error(
                explain2(exception, "[BADHEALTH] contacting server"), exc_info=exception
            )
            response = {"exception": str(exception)}

        return response

    # ----------------------------------------------------------------------------------------
    async def client_shutdown(self):
        """"""

        if not await self.__is_client_connection_possible():
            await self.close_client_session()
            return {"message": "no previous instance to shutdown"}

        await self._establish_client_session()

        time_zero = time.time()
        try:
            # TODO: Check response from shutdown, whatever it might be.
            await self._client_session.get(f"{self.__client_netloc}/shutdown")

            logger.debug(
                f"[PIDAL] client has requested {callsign(self)} server shutdown"
            )

            max_seconds = 5.0
            nap_seconds = 0.050
            timeout_time = time.time() + max_seconds

            while True:
                # Quit looping when server has closed its end of the socket.
                is_client_connection_possible = (
                    await self.__is_client_connection_possible()
                )
                if not is_client_connection_possible:
                    break

                # Too much time has passed while waiting for server to shut down?
                if time.time() > timeout_time:
                    raise RuntimeError(
                        f"remote server did not shut down within {max_seconds} seconds"
                    )

                logger.debug(
                    f"[PIDAL] waiting for {callsign(self)} server shutdown, napping {nap_seconds}"
                )

                # Small sleep after sending shutdown command lets the server really shut down.
                # TODO: Detect when server shutdown is complete.
                await asyncio.sleep(nap_seconds)
        finally:
            await self.close_client_session()

        return {
            # Let caller know that another process had existed.
            "had_existed": True,
            "message": "previous instance shutdown after %0.3f seconds"
            % (time.time() - time_zero),
        }

    # ----------------------------------------------------------------------------------------
    async def open_client_session(self):
        """"""

        await self._establish_client_session()

    # ----------------------------------------------------------------------------------------
    async def close_client_session(self):
        """"""

        if self._client_session is not None:
            # logger.debug("closing client session")

            await self._client_session.close()
            self._client_session = None



import logging

# Things created in the context.
from dls_servbase_api.datafaces.datafaces import (
    Datafaces,
    dls_servbase_datafaces_set_default,
)

logger = logging.getLogger(__name__)


class Context:
    """
    Client context for a dls_servbase_dataface object.
    On entering, it creates the object according to the specification (a dict).
    On exiting, it closes client connection.

    The aenter and aexit methods are exposed for use by an enclosing context.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        self.__specification = specification
        self.__dls_servbase_dataface = None

    # ----------------------------------------------------------------------------------------
    async def __aenter__(self):
        """ """

        await self.aenter()

    # ----------------------------------------------------------------------------------------
    async def __aexit__(self, type, value, traceback):
        """ """

        await self.aexit()

    # ----------------------------------------------------------------------------------------
    async def aenter(self):
        """ """

        # Build the object according to the specification.
        self.__dls_servbase_dataface = Datafaces().build_object(self.__specification)

        # If there is more than one dataface, the last one defined will be the default.
        dls_servbase_datafaces_set_default(self.__dls_servbase_dataface)

    # ----------------------------------------------------------------------------------------
    async def aexit(self):
        """ """

        if self.__dls_servbase_dataface is not None:
            await self.__dls_servbase_dataface.close_client_session()

            # Clear the global variable.  Important between pytests.
            dls_servbase_datafaces_set_default(None)



import asyncio
import logging
import multiprocessing

import pytest

from dls_servbase_api.databases.constants import CookieFieldnames, Tablenames
from dls_servbase_lib.databases.databases import Databases

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestDatabaseSqlite:
    def test(self, constants, logging_setup, output_directory):
        """
        Tests the sqlite implementation of the Database interface.

        This does not use a service.
        """

        database_specification = {
            "type": "dls_servbase_lib.databases.aiosqlite",
            "filename": "%s/dls_servbase_scheduler.sqlite" % (output_directory),
        }

        # Test direct SQL access to the database.
        DatabaseTesterImage().main(
            constants,
            database_specification,
            output_directory,
        )


# ----------------------------------------------------------------------------------------
class _BaseTester:
    """
    Provide asyncio loop and error checking over *Tester classes.
    """

    def main(self, constants, specification, output_directory):
        """
        This is the main program which calls the test using asyncio.
        """

        multiprocessing.current_process().name = "main"

        failure_message = None
        try:
            # Run main test in asyncio event loop.
            asyncio.run(
                self._main_coroutine(constants, specification, output_directory)
            )

        except Exception as exception:
            logger.exception(
                "unexpected exception in the test method", exc_info=exception
            )
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)


# ----------------------------------------------------------------------------------------
class DatabaseTesterImage(_BaseTester):
    """
    Test direct SQL access to the database.
    """

    async def _main_coroutine(
        self, constants, database_specification, output_directory
    ):
        """ """

        databases = Databases()
        database = databases.build_object(database_specification)

        try:
            # Connect to database.
            await database.connect()

            # Write one record.
            await database.insert(
                Tablenames.COOKIES,
                [
                    {
                        CookieFieldnames.UUID: "f0",
                        CookieFieldnames.CONTENTS: "{'a': 'f000'}",
                    }
                ],
            )
            all_sql = f"SELECT * FROM {Tablenames.COOKIES}"
            records = await database.query(all_sql)
            assert len(records) == 1

            # Bulk insert more records.
            insertable_records = [
                ["f1", "{'a': 'f111'}"],
                ["f2", "{'a': 'f111'}"],
                ["f3", "{'a': 'f111'}"],
                ["f4", "{'a': 'f111'}"],
            ]
            await database.execute(
                f"INSERT INTO {Tablenames.COOKIES}"
                f" ({CookieFieldnames.UUID}, {CookieFieldnames.CONTENTS})"
                " VALUES (?, ?)",
                insertable_records,
            )

            all_sql = f"SELECT * FROM {Tablenames.COOKIES}"
            records = await database.query(all_sql)
            assert len(records) == 5

        finally:
            # Connect from the database... necessary to allow asyncio loop to exit.
            await database.disconnect()



class Keywords:
    COMMAND = "dls_servbase_datafaces::keywords::command"
    PAYLOAD = "dls_servbase_datafaces::keywords::payload"


class Commands:
    EXECUTE = "dls_servbase_datafaces::commands::execute"



import logging

from dls_servbase_api.databases.constants import CookieFieldnames, Tablenames

# Object managing datafaces.
from dls_servbase_api.datafaces.datafaces import dls_servbase_datafaces_get_default

# Context creator.
from dls_servbase_lib.contexts.contexts import Contexts
from dls_servbase_lib.datafaces.context import Context as DatafaceContext

# Base class for the tester.
from tests.base_context_tester import BaseContextTester

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestDatafaceTakeover:
    def test_dataface_laptop(self, constants, logging_setup, output_directory):
        """ """

        configuration_file = "tests/configurations/servbase.yaml"
        DatafaceTakeoverTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class DatafaceTakeoverTester(BaseContextTester):
    """
    Class to test that a second instance of the dataface will cause the first one to shut down.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """

        dls_servbase_multiconf = self.get_multiconf()

        context_configuration = await dls_servbase_multiconf.load()
        dls_servbase_context = Contexts().build_object(context_configuration)

        async with dls_servbase_context:
            dataface = dls_servbase_datafaces_get_default()

            # Write one record.
            await dataface.insert(
                Tablenames.COOKIES,
                [
                    {
                        CookieFieldnames.UUID: "f0",
                        CookieFieldnames.CONTENTS: "{'a': 'f000'}",
                    }
                ],
            )

            # Make a new dataface context with the same specification.
            dataface_specification = dls_servbase_multiconf.require(
                "dls_servbase_dataface_specification"
            )

            # Activate the new dataface which should send shutdown to the old process.
            async with DatafaceContext(dataface_specification):
                all_sql = f"SELECT * FROM {Tablenames.COOKIES}"
                records = await dataface.query(all_sql)

                assert len(records) == 1
                assert records[0][CookieFieldnames.UUID] == "f0"
                assert records[0][CookieFieldnames.CONTENTS] == "{'a': 'f000'}"

            # From the top level context's point of view, one of its processs is dead.
            assert await dls_servbase_context.is_any_process_dead()

        assert not await dls_servbase_context.is_any_process_alive()



import logging
import os
import tempfile

from dls_multiconf_lib.multiconfs import Multiconfs, multiconfs_set_default

# Utilities.
from dls_utilpack.visit import get_visit_year

logger = logging.getLogger(__name__)


class Base:
    """
    Base class for femtocheck subcommands.  Handles details like configuration.
    """

    def __init__(self, args):
        self._args = args

        self.__temporary_directory = None

    # ----------------------------------------------------------------------------------------
    def get_multiconf(self):

        dls_servbase_multiconf = Multiconfs().build_object_from_environment()

        # For convenience, make a temporary directory for this test.
        self.__temporary_directory = tempfile.TemporaryDirectory()

        # Make the temporary directory available to the multiconf.
        dls_servbase_multiconf.substitute(
            {"temporary_directory": self.__temporary_directory.name}
        )

        substitutions = {
            "APPS": "/dls_sw/apps",
            "CWD": os.getcwd(),
            "HOME": os.environ.get("HOME", "HOME"),
            # Provide the PYTHONPATH at the time of service start
            # to the (potentially remote) process where the dls_servbase_task.run is called.
            "PYTHONPATH": os.environ.get("PYTHONPATH", "PYTHONPATH"),
        }

        if hasattr(self._args, "visit") and self._args.visit != "VISIT":
            BEAMLINE = os.environ.get("BEAMLINE")
            if BEAMLINE is None:
                raise RuntimeError("BEAMLINE environment variable is not defined")
            year = get_visit_year(BEAMLINE, self._args.visit)
            substitutions["BEAMLINE"] = BEAMLINE
            substitutions["VISIT"] = self._args.visit
            substitutions["YEAR"] = year

        dls_servbase_multiconf.substitute(substitutions)

        # Set this as the default multiconf so it is available everywhere.
        multiconfs_set_default(dls_servbase_multiconf)

        return dls_servbase_multiconf



import logging

# Base class for table definitions.
from dls_normsql.table_definition import TableDefinition

from dls_servbase_api.databases.constants import CookieFieldnames, Tablenames

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class CookiesTableDefinition(TableDefinition):
    # ----------------------------------------------------------------------------------------
    def __init__(self):
        TableDefinition.__init__(self, Tablenames.COOKIES)

        # All cookies have a unique uuid field.
        self.fields[CookieFieldnames.UUID] = {
            "type": "TEXT PRIMARY KEY",
            "index": True,
        }

        self.fields[CookieFieldnames.CREATED_ON] = {"type": "TEXT", "index": True}
        self.fields[CookieFieldnames.NAME] = {"type": "TEXT", "index": True}
        self.fields[CookieFieldnames.CONTENTS] = {"type": "TEXT", "index": False}



import logging

from dls_servbase_api.databases.constants import CookieFieldnames, Tablenames

# Base class for all aiosqlite database objects.
from dls_servbase_lib.databases.table_definitions import CookiesTableDefinition

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class DatabaseDefinition:
    """
    Class which defines the database tables and revision migration path.
    Used in concert with the normsql class.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self):
        """
        Construct object.  Do not connect to database.
        """

        self.LATEST_REVISION = 2

    # ----------------------------------------------------------------------------------------
    async def apply_revision(self, revision):
        if revision == 2:
            await self.execute(
                f"ALTER TABLE {Tablenames.COOKIES} ADD COLUMN {CookieFieldnames.NAME} TEXT",
                why=f"revision 2: add {Tablenames.COOKIES} {CookieFieldnames.NAME} column",
            )
            await self.execute(
                "CREATE INDEX %s_%s ON %s(%s)"
                % (
                    Tablenames.COOKIES,
                    CookieFieldnames.NAME,
                    Tablenames.COOKIES,
                    CookieFieldnames.NAME,
                )
            )

    # ----------------------------------------------------------------------------------------
    async def add_table_definitions(self):
        """
        Make all the table definitions.
        """

        # Table schemas in our database.
        self.add_table_definition(CookiesTableDefinition())



import logging

from dls_utilpack.callsign import callsign

# Utilities.
from dls_utilpack.explain import explain

# Base class which maps flask requests to methods.
from dls_servbase_lib.contexts.base import Base

# Contexts.
from dls_servbase_lib.datafaces.context import Context as DatafaceContext
from dls_servbase_lib.guis.context import Context as GuiContext

logger = logging.getLogger(__name__)


thing_type = "dls_servbase_lib.dls_servbase_contexts.classic"


class Classic(Base):
    """
    Object representing an event dls_servbase_dataface connection.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        Base.__init__(self, thing_type, specification)

        self.__dataface = None
        self.__collector = None
        self.__gui = None

    # ----------------------------------------------------------------------------------------
    async def __dead_or_alive(self, server, dead, alive):

        if server is not None:
            # A server was defined for this context?
            if await server.is_process_started():
                if await server.is_process_alive():
                    alive.append(server)
                else:
                    dead.append(server)

    # ----------------------------------------------------------------------------------------
    async def __dead_or_alive_all(self):
        """
        Return two lists, one for dead and one for alive processes.
        TODO: Parallelize context process alive/dead checking.
        """

        dead = []
        alive = []

        await self.__dead_or_alive(self.__dataface, dead, alive)
        await self.__dead_or_alive(self.__collector, dead, alive)
        await self.__dead_or_alive(self.__gui, dead, alive)

        return dead, alive

    # ----------------------------------------------------------------------------------------
    async def is_any_process_alive(self):
        """
        Check all configured processes, return if any alive.
        """
        dead, alive = await self.__dead_or_alive_all()

        logger.debug(f"{len(dead)} processes are dead, {len(alive)} are alive")

        return len(alive) > 0

    # ----------------------------------------------------------------------------------------
    async def is_any_process_dead(self):
        """
        Check all configured processes, return if any alive.
        """
        dead, alive = await self.__dead_or_alive_all()

        return len(dead) > 0

    # ----------------------------------------------------------------------------------------
    async def __aenter__(self):
        """ """
        logger.debug(f"entering {callsign(self)} context")

        try:

            try:
                specification = self.specification().get(
                    "dls_servbase_dataface_specification"
                )
                if specification is not None:
                    logger.debug(f"at entering position {callsign(self)} DATAFACE")
                    self.__dataface = DatafaceContext(specification)
                    await self.__dataface.aenter()
            except Exception as exception:
                raise RuntimeError(
                    explain(exception, f"creating {callsign(self)} dataface context")
                )

            try:
                specification = self.specification().get(
                    "dls_servbase_gui_specification"
                )
                if specification is not None:
                    logger.debug(f"at entering position {callsign(self)} GUI")
                    self.__gui = GuiContext(specification)
                    await self.__gui.aenter()
            except Exception as exception:
                raise RuntimeError(
                    explain(exception, f"creating {callsign(self)} gui context")
                )

        except Exception as exception:
            await self.aexit()
            raise RuntimeError(explain(exception, f"entering {callsign(self)} context"))

        logger.debug(f"entered {callsign(self)} context")

    # ----------------------------------------------------------------------------------------
    async def __aexit__(self, type, value, traceback):
        """ """

        await self.aexit()

    # ----------------------------------------------------------------------------------------
    async def aexit(self):
        """ """

        logger.debug(f"exiting {callsign(self)} context")

        if self.__gui is not None:
            logger.debug(f"at exiting position {callsign(self)} GUI")
            try:
                await self.__gui.aexit()
            except Exception as exception:
                logger.error(
                    explain(exception, f"exiting {callsign(self.__gui)} context"),
                    exc_info=exception,
                )
            self.__gui = None

        if self.__dataface is not None:
            logger.debug(f"at exiting position {callsign(self)} DATAFACE")
            try:
                await self.__dataface.aexit()
            except Exception as exception:
                logger.error(
                    explain(exception, f"exiting {callsign(self.__dataface)} context"),
                    exc_info=exception,
                )
            self.__dataface = None

        logger.debug(f"exited {callsign(self)} context")



import logging

# Utilities.
from dls_utilpack.callsign import callsign

# Base class for generic things.
from dls_utilpack.thing import Thing

logger = logging.getLogger(__name__)


class Base(Thing):
    """ """

    # ----------------------------------------------------------------------------------------
    def __init__(self, thing_type, specification=None, predefined_uuid=None):
        Thing.__init__(self, thing_type, specification, predefined_uuid=predefined_uuid)

        self.__contents = None
        self.__is_dirty = False

    # -----------------------------------------------------------------------------
    def set_dirty(self, dirty=True):
        self.__is_dirty = dirty

    def is_dirty(self):
        return self.__is_dirty

    def set_contents(self, contents):
        self.__contents = contents

    def get_contents(self):
        if self.__contents is None:
            raise RuntimeError(f"{callsign(self)} contents has not been set")
        return self.__contents

    def has_contents(self):
        return self.__contents is not None



import logging

# Base class for generic things.
from dls_utilpack.thing import Thing

from dls_servbase_api.databases.constants import Tablenames

# Database manager.
from dls_servbase_lib.databases.databases import Databases

logger = logging.getLogger(__name__)

thing_type = "dls_servbase_lib.datafaces.aiosqlite"


class Aiosqlite(Thing):
    """ """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):
        Thing.__init__(self, thing_type, specification)

        self.__database = None

    # ----------------------------------------------------------------------------------------
    async def start(self):
        # Connect to the database to create the schemas if they don't exist already.
        await self.establish_database_connection()

    # ----------------------------------------------------------------------------------------
    async def disconnect(self):
        if self.__database is not None:
            await self.__database.disconnect()
            self.__database = None

    # ----------------------------------------------------------------------------------------
    async def establish_database_connection(self):

        if self.__database is None:
            self.__database = Databases().build_object(self.specification()["database"])
            await self.__database.connect()

    # ----------------------------------------------------------------------------------------
    async def reinstance(self):
        """"""
        if self.__database is None:
            return

        self.__database = self.__database.reinstance()

    # ----------------------------------------------------------------------------------------
    async def set_cookie(self, record):
        """ """
        await self.establish_database_connection()

        await self.__database.insert(Tablenames.COOKIES, [record])

    # ----------------------------------------------------------------------------------------
    async def get_cookie(self, cookie_uuid):
        """
        Get single cookie from its uuid.
        Returns database record format.
        """
        await self.establish_database_connection()

        sql = "SELECT * FROM cookies WHERE uuid = '%s'" % (cookie_uuid)

        records = await self.__database.query(sql, why="[COOKSEL]")

        if len(records) == 0:
            return None

        return records[0]

    # ----------------------------------------------------------------------------------------
    async def update_cookie(self, row):
        """"""
        await self.establish_database_connection()

        count = await self.__database.update(
            Tablenames.COOKIES, row, "uuid = '%s'" % (row["uuid"]), why="[COOKSEL]"
        )

        return {"count": count}

    # ----------------------------------------------------------------------------------------
    async def backup(self):
        """"""
        await self.establish_database_connection()

        return await self.__database.backup()

    # ----------------------------------------------------------------------------------------
    async def restore(self, nth):
        """"""
        await self.establish_database_connection()

        return await self.__database.restore(nth)

    # ----------------------------------------------------------------------------------------
    async def query(self, sql, subs=None, why=None):
        """"""
        await self.establish_database_connection()

        records = await self.__database.query(sql, subs=subs, why=why)

        return records

    # ----------------------------------------------------------------------------------------
    async def execute(self, sql, subs=None, why=None):
        """"""
        await self.establish_database_connection()

        return await self.__database.execute(sql, subs=subs, why=why)

    # ----------------------------------------------------------------------------------------
    async def insert(self, table_name, records, why=None):
        """"""
        await self.establish_database_connection()

        if why is None:
            why = f"insert {len(records)} {table_name} records"

        await self.__database.insert(table_name, records, why=why)

    # ----------------------------------------------------------------------------------------
    async def update(self, table_name, record, where, subs=None, why=None):
        """"""
        await self.establish_database_connection()

        if why is None:
            why = f"update {table_name} record"

        # This returns the count of records changed by the update.
        return {
            "count": await self.__database.update(
                table_name, record, where, subs=subs, why=why
            )
        }

    # ----------------------------------------------------------------------------------------
    async def report_health(self):
        """"""

        report = {}

        report["alive"] = True

        return report



import logging

from dls_servbase_api.databases.constants import CookieFieldnames, Tablenames

# Object managing datafaces.
from dls_servbase_api.datafaces.datafaces import dls_servbase_datafaces_get_default

# Context creator.
from dls_servbase_lib.contexts.contexts import Contexts

# Base class for the tester.
from tests.base_context_tester import BaseContextTester

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestDataface:
    """
    Test that we can do a basic database operation through the service.
    """

    def test(self, constants, logging_setup, output_directory):
        """ """

        configuration_file = "tests/configurations/servbase.yaml"
        DatafaceTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class DatafaceTester(BaseContextTester):
    """
    Class to test the dataface.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """

        dls_servbase_multiconf = self.get_multiconf()

        context_configuration = await dls_servbase_multiconf.load()

        dls_servbase_context = Contexts().build_object(context_configuration)

        async with dls_servbase_context:
            dataface = dls_servbase_datafaces_get_default()

            # Write one record.
            await dataface.insert(
                Tablenames.COOKIES,
                [
                    {
                        CookieFieldnames.UUID: "f0",
                        CookieFieldnames.CONTENTS: "{'a': 'f000'}",
                    }
                ],
            )

            all_sql = f"SELECT * FROM {Tablenames.COOKIES}"
            records = await dataface.query(all_sql)

            assert len(records) == 1
            assert records[0][CookieFieldnames.UUID] == "f0"
            assert records[0][CookieFieldnames.CONTENTS] == "{'a': 'f000'}"

            # ----------------------------------------------------------------
            # Now try a direct update.
            record = {
                CookieFieldnames.CONTENTS: "{'b': 'f1111'}",
            }

            subs = ["f0"]
            result = await dataface.update(
                Tablenames.COOKIES,
                record,
                f"{CookieFieldnames.UUID} = ?",
                subs=subs,
                why="test update",
            )

            assert result["count"] == 1
            records = await dataface.query(all_sql)

            assert len(records) == 1
            assert records[0][CookieFieldnames.UUID] == "f0"
            assert records[0][CookieFieldnames.CONTENTS] == "{'b': 'f1111'}"

            # ----------------------------------------------------------------
            # Now try a high level API update.
            record = {
                CookieFieldnames.UUID: "f0",
                CookieFieldnames.CONTENTS: "{'c': 'f2222'}",
            }

            result = await dataface.update_cookie(
                record,
            )

            assert result["count"] == 1
            records = await dataface.query(all_sql)

            assert len(records) == 1
            assert records[0][CookieFieldnames.UUID] == "f0"
            assert records[0][CookieFieldnames.CONTENTS] == "{'c': 'f2222'}"



import logging

# Utilities.
from dls_utilpack.describe import describe

# API constants.
from dls_servbase_api.constants import Keywords as ProtocoljKeywords

# Context creator.
from dls_servbase_lib.contexts.contexts import Contexts
from dls_servbase_lib.guis.constants import Commands, Cookies, Keywords

# Object managing gui
from dls_servbase_lib.guis.guis import dls_servbase_guis_get_default

# Base class for the tester.
from tests.base_context_tester import BaseContextTester

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestGui:
    """
    Test the dls_servbase gui service.

    Fetch requests from the dataface similar to what javascript ajax would do.
    """

    def test(self, constants, logging_setup, output_directory):
        """ """

        configuration_file = "tests/configurations/servbase.yaml"
        GuiTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class GuiTester(BaseContextTester):
    """
    Class with asyncio coroutine which does the actual test.
    """

    # Called from the base class main method.
    async def _main_coroutine(self, constants, output_directory):
        """ """

        # Make a multiconf and load the context configuration file.
        multiconf = self.get_multiconf()
        context_configuration = await multiconf.load()

        dls_servbase_gui_specification = context_configuration[
            "dls_servbase_gui_specification"
        ]
        type_specific_tbd = dls_servbase_gui_specification["type_specific_tbd"]
        aiohttp_specification = type_specific_tbd["aiohttp_specification"]
        aiohttp_specification["search_paths"] = [output_directory]

        # Make a context instance which can start a service and stop it at the end.
        dls_servbase_context = Contexts().build_object(context_configuration)

        async with dls_servbase_context:

            # --------------------------------------------------------------------
            # Use protocolj to fetch a request from the dataface.
            # Simulates what javascript would do by ajax.

            # Load tabs, of which there are none at the start.
            # This establishes the cookie though.
            load_tabs_request = {
                Keywords.COMMAND: Commands.LOAD_TABS,
                ProtocoljKeywords.ENABLE_COOKIES: [Cookies.TABS_MANAGER],
            }

            response = await dls_servbase_guis_get_default().client_protocolj(
                load_tabs_request, cookies={}
            )

            # The response is json, with the last saved tab_id, which is None at first.
            assert Keywords.TAB_ID in response
            assert response[Keywords.TAB_ID] is None

            # We should also have cookies back.
            assert "__cookies" in response
            cookies = response["__cookies"]
            assert Cookies.TABS_MANAGER in cookies

            # Use the cookie name in the next requests.
            cookie_uuid = cookies[Cookies.TABS_MANAGER]

            # --------------------------------------------------------------------
            # Select a tab.
            # The response is json, but nothing really to see in it.

            select_tab_request = {
                Keywords.COMMAND: Commands.SELECT_TAB,
                ProtocoljKeywords.ENABLE_COOKIES: [Cookies.TABS_MANAGER],
                Keywords.TAB_ID: "123",
            }

            response = await dls_servbase_guis_get_default().client_protocolj(
                select_tab_request, cookies={Cookies.TABS_MANAGER: cookie_uuid}
            )

            # --------------------------------------------------------------------
            # Load tabs again, this time we should get the saved tab_id.

            response = await dls_servbase_guis_get_default().client_protocolj(
                load_tabs_request, cookies={Cookies.TABS_MANAGER: cookie_uuid}
            )

            logger.debug(describe("second load_tabs response", response))
            assert Keywords.TAB_ID in response
            assert response[Keywords.TAB_ID] == "123"

            # --------------------------------------------------------------------
            # Update a tab.
            # The response is json, but nothing really to see in it.

            select_tab_request = {
                Keywords.COMMAND: Commands.SELECT_TAB,
                ProtocoljKeywords.ENABLE_COOKIES: [Cookies.TABS_MANAGER],
                Keywords.TAB_ID: "456",
            }

            response = await dls_servbase_guis_get_default().client_protocolj(
                select_tab_request, cookies={Cookies.TABS_MANAGER: cookie_uuid}
            )

            # --------------------------------------------------------------------
            # Write a text file and fetch it through the http server.
            filename = "test.html"
            contents = "some test html"
            with open(f"{output_directory}/{filename}", "wt") as file:
                file.write(contents)
            text = await dls_servbase_guis_get_default().client_get_file(filename)
            # assert text == contents

            # Write a binary file and fetch it through the http server.
            filename = "test.exe"
            contents = "some test binary"
            with open(f"{output_directory}/{filename}", "wt") as file:
                file.write(contents)
            binary = await dls_servbase_guis_get_default().client_get_file(filename)
            # Binary comes back as bytes due to suffix of url filename.
            assert binary == contents.encode()

            # --------------------------------------------------------------------
            # Get an html file automatically configured in guis/html.
            filename = "javascript/version.js"
            # TODO: Put proper version into javascript somehow during packaging.
            contents = "dls_servbase___CURRENT_VERSION"
            text = await dls_servbase_guis_get_default().client_get_file(filename)
            logger.debug(f"javascript version is {text.strip()}")
            assert contents in text



import logging

# Base class which maps flask requests to methods.
from dls_servbase_lib.contexts.base import Base as ContextBase

# Things created in the context.
from dls_servbase_lib.guis.guis import Guis, dls_servbase_guis_set_default

logger = logging.getLogger(__name__)


thing_type = "dls_servbase_lib.dls_servbase_guis.context"


class Context(ContextBase):
    """
    Object representing an event dls_servbase_dataface connection.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        ContextBase.__init__(self, thing_type, specification)

    # ----------------------------------------------------------------------------------------
    async def aenter(self):
        """ """

        self.server = Guis().build_object(self.specification())

        # If there is more than one gui, the last one defined will be the default.
        dls_servbase_guis_set_default(self.server)

        if self.context_specification.get("start_as") == "coro":
            await self.server.activate_coro()

        elif self.context_specification.get("start_as") == "thread":
            await self.server.start_thread()

        elif self.context_specification.get("start_as") == "process":
            await self.server.start_process()

    # ----------------------------------------------------------------------------------------
    async def aexit(self):
        """ """

        if self.server is not None:
            # Put in request to shutdown the server.
            await self.server.client_shutdown()

            # Release a client connection if we had one.
            await self.server.close_client_session()

        dls_servbase_guis_set_default(None)



import asyncio
import concurrent.futures
import functools
import logging
import multiprocessing
from urllib.parse import urlparse

import aiohttp
import aiohttp.web
import aiohttp.web_runner
import pytest

# Utilities.
from dls_utilpack.describe import describe

logger = logging.getLogger(__name__)


class TestAiohttp:
    """
    Test the aiohttp package directly, just to make sure it operations like we think it does.

    Does not import from any of the dls_servbase packages.

    Start a process for a server, and one for a client, and let them talk.

    """

    # ----------------------------------------------------------------------------------------
    def test_internet_sockest(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        multiprocessing.current_process().name = "main"

        failure_message = None

        try:
            # Run main in asyncio event loop.
            asyncio.run(
                self._main_coroutine(
                    constants,
                    "http://*:22132",
                    "http://localhost:22132",
                    output_directory,
                )
            )

        except Exception as exception:
            logger.exception("unexpected exception during the test", exc_info=exception)
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)

    # ----------------------------------------------------------------------------------------
    def test_unix_pipes(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        multiprocessing.current_process().name = "main"

        failure_message = None

        try:
            # Run main in asyncio event loop.
            asyncio.run(
                self._main_coroutine(
                    constants,
                    f"{output_directory}/unix_socket",
                    f"{output_directory}/unix_socket",
                    output_directory,
                )
            )

        except Exception as exception:
            logger.exception("unexpected exception during the test", exc_info=exception)
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)

    # ----------------------------------------------------------------------------------------
    async def _main_coroutine(
        self, constants, server_endpoint, client_endpoint, output_directory
    ):
        """ """

        loop = asyncio.get_running_loop()

        self.__server_endpoint = server_endpoint
        self.__client_endpoint = client_endpoint

        with concurrent.futures.ProcessPoolExecutor() as pool:
            # Start a server and client in separate processes.
            server_executor = loop.run_in_executor(
                pool,
                functools.partial(self._server_process, constants, output_directory),
            )

            # Let the server process get truly started.
            # TODO: Remove sleep waiting for server process to start in test_31_aiohttp.
            await asyncio.sleep(0.5)
            client_executor = loop.run_in_executor(
                pool,
                functools.partial(self._client_process, constants, output_directory),
            )

            client_result = await client_executor
            logger.info(describe("client_result", client_result))

            server_result = await server_executor
            logger.info(describe("server_result", server_result))

    # ----------------------------------------------------------------------------------------
    def _server_process(
        self,
        constants,
        output_directory,
    ):
        """ """
        multiprocessing.current_process().name = "_server"

        try:
            logger.info("in _server_process")
            # raise RuntimeError("deliberate exception in _server_process")
            server_task = self._server_coroutine(constants, output_directory)
            server_task_result = asyncio.run(server_task)
            return server_task_result

        except Exception as exception:
            logger.exception("error in server process", exc_info=exception)
            return exception

    # ----------------------------------------------------------------------------------------
    async def _server_coroutine(
        self,
        constants,
        output_directory,
    ):
        """ """

        self._shutdown_event = asyncio.Event()

        try:
            logger.info("in _server_coroutine")
            web_app = aiohttp.web.Application()
            web_app.add_routes(
                [
                    aiohttp.web.get("/hello", self._handle_hello),
                    aiohttp.web.get("/shutdown", self._handle_shutdown),
                ]
            )

            self._runner = aiohttp.web.AppRunner(web_app)
            await self._runner.setup()

            if self.__server_endpoint.startswith("http"):
                parts = urlparse(self.__server_endpoint)
                site = aiohttp.web.TCPSite(self._runner, parts.hostname, parts.port)
            else:
                site = aiohttp.web.UnixSite(self._runner, self.__server_endpoint)
            await site.start()

            logger.info("awaiting shutdown event")
            await self._shutdown_event.wait()

            return None

        except Exception as exception:
            logger.exception("error in server coroutine", exc_info=exception)
            return exception

    # ----------------------------------------------------------------------------------------
    async def _handle_hello(self, request):
        try:
            logger.info("handling hello")
            # raise RuntimeError("deliberate exception in _handle_hello")
            response = aiohttp.web.Response(text="Hello, world")
            return response
        except Exception as exception:
            return self._componse_exception(exception)

    # ----------------------------------------------------------------------------------------
    async def _handle_shutdown(self, request):
        try:
            logger.info("handling shutdown")
            response = aiohttp.web.Response(status=200, text="ok")
            logger.info("setting shutdown event")
            self._shutdown_event.set()
            return response
        except Exception as exception:
            return self._componse_exception(exception)

    # ----------------------------------------------------------------------------------------
    def _componse_exception(self, exception):
        logger.exception("error ", exc_info=exception)
        return aiohttp.web.Response(status=500, body=str(exception).encode())

    # ----------------------------------------------------------------------------------------
    def _client_process(
        self,
        constants,
        output_directory,
    ):
        """ """
        multiprocessing.current_process().name = "_client"

        try:
            logger.info("in _client_process")

            # raise RuntimeError("deliberate exception in _client_process")
            client_task = self._client_coroutine(constants, output_directory)
            client_task_result = asyncio.run(client_task)
            return client_task_result

        except Exception as exception:
            logger.exception("error in client process", exc_info=exception)
            return exception

    # ----------------------------------------------------------------------------------------
    async def _client_coroutine(
        self,
        constants,
        output_directory,
    ):
        """ """

        try:
            logger.info("in _client_coroutine")
            # raise RuntimeError("deliberate exception in _client_coroutine")

            if self.__client_endpoint.startswith("http"):
                connector = None
                netloc = self.__client_endpoint
            else:
                connector = aiohttp.UnixConnector(path=self.__client_endpoint)
                netloc = "http://unixconnector"

            logger.info(f"client netloc is {netloc}")

            async with aiohttp.ClientSession(connector=connector) as session:
                try:
                    async with session.get(f"{netloc}/hello") as hello_response:
                        pass
                except Exception:
                    logger.exception("error in client call")

                async with session.get(f"{netloc}/shutdown"):
                    pass
                return hello_response.status
                # print(await resp.text())

        except Exception as exception:
            logger.exception("error in client coroutine", exc_info=exception)
            return exception
